/**
 * @file
 * @brief CAN driver thread.
 */

#include <sta/config.hpp>
#ifdef STA_CAN_BUS_ENABLE

#include <sta/debug/assert.hpp>
#include <sta/debug/debug.hpp>
#include <sta/bus/can/subscribable.hpp>
#include <sta/lang.hpp>
#include <sta/rtos/defs.hpp>
#include <sta/rtos/event.hpp>
#include <sta/devices/stm32/hal.hpp>
#include <sta/rtos/system/can_bus.hpp>

#include <cmsis_os2.h>
#include <FreeRTOS.h>

#include <cstring>

namespace debug
{
	/**
	 * @brief Output CAN frame ID to UART.
	 *
	 * @param id    Frame ID
	 */
	void printFrameID(const sta::CanFrameId & id)
	{
		STA_DEBUG_PRINT("SID: ");
		STA_DEBUG_PRINTLN(id.sid, sta::IntegerBase::HEX);
		if (id.format == sta::CanIdFormat::EXT)
		{
			STA_DEBUG_PRINT("EID: ");
			STA_DEBUG_PRINTLN(id.eid, sta::IntegerBase::HEX);
		}
	}

	/**
	 * @brief Output CAN frame payload to UART.
	 *
	 * @param payload   Payload buffer
	 * @param size      Payload size
	 */
	void printPayloadHex(const uint8_t * payload, uint8_t size)
	{
		// Write frame payload to DebugSerial
		STA_DEBUG_PRINT("payload: ");
		for (uint8_t i = 0; i < size; ++i)
		{
			STA_DEBUG_PRINT(payload[i], sta::IntegerBase::HEX);
			STA_DEBUG_PRINT(' ');
		}
		STA_DEBUG_PRINTLN();
	}
} // namespace debug

namespace sta
{
	AlpakaCanBus::AlpakaCanBus(CanController * controller, TimeMsFn timeMs)
		: controller_{controller}
	{
		setupSubscriptions();
	}


	void AlpakaCanBus::send(const CanSysMsg & msg)
	{
		CanTxHeader header;
		header.id.format     = static_cast<CanIdFormat>(msg.header.format);
		header.id.sid        = msg.header.sid;
		header.id.eid        = msg.header.eid;
		header.payloadLength = msg.header.payloadLength;

		controller_->sendFrame(header, msg.payload);
	}

	void AlpakaCanBus::setupSubscriptions()
	{
		// Make sure to receive all messages
		CanFilter filter;

		// Clear previous subscriptions
		controller_->clearFilters();


		// All bits except [0:1] of the SID must be zero for system messages
		filter.obj  = {0, 0};
		filter.mask = {~STA_CAN_SID_SYS_BITS, 0};
		filter.type = sta::CanFilterIdFormat::ANY;
		filter.fifo = FIFO_SYS;
		controller_->configureFilter(FIFO_SYS, filter, true);

		// TODO Limit which data messages are received
		// Bits [0:1] of the SID must be zero for data messages
		filter.obj  = {0, 0};
		filter.mask = {STA_CAN_SID_SYS_BITS, 0};
		filter.type = sta::CanFilterIdFormat::ANY;
		filter.fifo = FIFO_DATA;
		controller_->configureFilter(FIFO_DATA, filter, true);
	}
} // namespace sta

#endif // STA_CAN_BUS_ENABLE
